@@ -15,9 +15,6 @@
 #include "libsandbox.h"
 #include "sbutil.h"
 
-/* Pick a value to guarantee alignment requirements. #565630 */
-#define MIN_ALIGN (2 * sizeof(void *))
-
 /* Well screw me sideways, someone decided to override mmap() #290249
  * We probably don't need to include the exact sym version ...
  */
@@ -33,6 +30,7 @@ static void *sb_mmap(void *addr, size_t length, int prot, int flags, int fd, off
 	return _sb_mmap(addr, length, prot, flags, fd, offset);
 }
 #define mmap sb_mmap
+
 static int (*_sb_munmap)(void *addr, size_t length);
 static int sb_munmap(void *addr, size_t length)
 {
@@ -42,27 +40,46 @@ static int sb_munmap(void *addr, size_t length)
 }
 #define munmap sb_munmap
 
-#define SB_MALLOC_TO_MMAP(ptr) ((void*)((uintptr_t)(ptr) - MIN_ALIGN))
-#define SB_MMAP_TO_MALLOC(ptr) ((void*)((uintptr_t)(ptr) + MIN_ALIGN))
-#define SB_MALLOC_TO_MMAP_SIZE(ptr) (*((size_t*)SB_MALLOC_TO_MMAP(ptr)))
-#define SB_MALLOC_TO_SIZE(ptr) (SB_MALLOC_TO_MMAP_SIZE(ptr) - MIN_ALIGN)
+static void *(*_sb_mremap)(void *old_address, size_t old_size, size_t new_size, int flags);
+static void *sb_mremap(void *old_address, size_t old_size, size_t new_size, int flags)
+{
+	if (!_sb_mremap)
+		_sb_mremap = get_dlsym("mremap", NULL);
+	return _sb_mremap(old_address, old_size, new_size, flags);
+}
+#define mremap sb_mremap
+
+/* Ensure malloc returns aligned memory #565630 */
+#define ALIGN_FACTOR 2
+#define ALIGN_SIZE (ALIGN_FACTOR * sizeof(size_t))
 
 void *malloc(size_t size)
 {
-	size_t *ret;
-	size += MIN_ALIGN;
-	ret = mmap(0, size, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
-	if (ret == MAP_FAILED)
+	long pagesize = sysconf(_SC_PAGESIZE);
+
+	if (__builtin_add_overflow(size, ALIGN_SIZE, &size) ||
+			__builtin_add_overflow(size, pagesize - size % pagesize, &size)) {
+		errno = ENOMEM;
 		return NULL;
-	*ret = size;
-	return SB_MMAP_TO_MALLOC(ret);
+	}
+
+	size_t *sp = mmap(0, size, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
+	if (sp == MAP_FAILED)
+		return NULL;
+
+	*sp = size;
+	return sp + ALIGN_FACTOR;
 }
 
 void free(void *ptr)
 {
 	if (ptr == NULL)
 		return;
-	if (munmap(SB_MALLOC_TO_MMAP(ptr), SB_MALLOC_TO_MMAP_SIZE(ptr)))
+
+	size_t *sp = ptr;
+	sp -= ALIGN_FACTOR;
+
+	if (munmap(sp, *sp))
 		sb_ebort("sandbox memory corruption with free(%p): %s\n",
 			ptr, strerror(errno));
 }
@@ -70,17 +87,20 @@ void free(void *ptr)
 /* Hrm, implement a zalloc() ? */
 void *calloc(size_t nmemb, size_t size)
 {
-	void *ret = malloc(nmemb * size); /* dont care about overflow */
+	if (__builtin_mul_overflow(nmemb, size, &size)) {
+		errno = ENOMEM;
+		return NULL;
+	}
+
+	void *ret = malloc(size);
 	if (ret)
-		memset(ret, 0, nmemb * size);
+		memset(ret, 0, size);
+
 	return ret;
 }
 
 void *realloc(void *ptr, size_t size)
 {
-	void *ret;
-	size_t old_malloc_size;
-
 	if (ptr == NULL)
 		return malloc(size);
 	if (size == 0) {
@@ -88,20 +108,34 @@ void *realloc(void *ptr, size_t size)
 		return NULL;
 	}
 
-	old_malloc_size = SB_MALLOC_TO_SIZE(ptr);
-	/* Since mmap() is heavy, don't bother shrinking */
-	if (size <= old_malloc_size)
-		return ptr;
-	ret = malloc(size);
-	if (!ret)
-		return ret;
-	/* We already verified old_malloc_size is smaller than size above, so
-	 * we don't really need the MIN() here.  We leave it to be defensive,
-	 * and because gcc optimizes away the check for us.
-	 */
-	memcpy(ret, ptr, MIN(size, old_malloc_size));
-	free(ptr);
-	return ret;
+	long pagesize = sysconf(_SC_PAGESIZE);
+
+	if (__builtin_add_overflow(size, ALIGN_SIZE, &size) ||
+			__builtin_add_overflow(size, pagesize - size % pagesize, &size)) {
+		errno = ENOMEM;
+		return NULL;
+	}
+
+	size_t *sp = ptr;
+	sp -= 2;
+
+	sp = mremap(sp, *sp, size, MREMAP_MAYMOVE);
+	if (sp == MAP_FAILED)
+		return NULL;
+
+	*sp = size;
+	return sp + ALIGN_FACTOR;
+}
+
+size_t malloc_size(void *ptr)
+{
+	if (!ptr)
+		return 0;
+
+	size_t *sp = ptr;
+	sp -= 2;
+
+	return *sp - ALIGN_SIZE;
 }
 
 char *strdup(const char *s)
