@@ -5,6 +5,4 @@
  * Licensed under the GPL-2
  */
 
-#define sb_openat_pre_check sb_openat64_pre_check
 #include "__open_2.c"
-#undef sb_openat_pre_check
