@@ -115,16 +115,25 @@ typedef struct user_regs_struct trace_regs;
 # define AT_SYMLINK_NOFOLLOW 0
 #endif
 
-#if !HAVE_DLVSYM
-# define dlvsym(_lib, _sym, _ver) dlsym(_lib, _sym)
-#endif
-
-#if HAVE_DLVSYM
+#ifdef HAVE_DLVSYM
 /* Taken from glibc */
 # define symbol_version(_real, _name, _version) \
 	__asm__ (".symver " #_real "," #_name "@" #_version)
 # define default_symbol_version(_real, _name, _version) \
 	__asm__ (".symver " #_real "," #_name "@@" #_version)
+#else
+# define dlvsym(_lib, _sym, _ver) dlsym(_lib, _sym)
+#endif
+
+#ifndef HAVE_STRLCPY
+static size_t strlcpy(char *dst, const char *src, size_t size)
+{
+	size_t srclen = strlen(src);
+	size_t copylen = (srclen < size ? srclen : size - 1);
+	memcpy(dst, src, copylen);
+	dst[copylen] = '\0';
+	return srclen;
+}
 #endif
 
 /* Taken from glibc */
